import * as babylon from 'babylon'
import {
  FunctionExpression
, ArrowFunctionExpression
, ObjectExpression
, ObjectMethod
, StringLiteral
, Identifier
, NumericLiteral
, TemplateLiteral
} from 'babel-types'
import generate from '@babel/generator'

describe('Basic', () => {
  test('function() {}', () => {
    const result = babylon.parseExpression('function() {}') as FunctionExpression
    expect(result.type).toEqual('FunctionExpression')
    expect(result.generator).toBeFalsy()
    expect(result.async).toBeFalsy()
    expect(result.params.length).toEqual(0)
    const { code } = generate(result)
    expect(code).toEqual('function () {}')
  })

  test('() => {}', () => {
    const result = babylon.parseExpression('function() {}') as FunctionExpression
    expect(result.type).toEqual('FunctionExpression')
    expect(result.generator).toBeFalsy()
    expect(result.async).toBeFalsy()
    expect(result.params.length).toEqual(0)
    const { code } = generate(result)
    expect(code).toEqual('function () {}')
  })
})

describe('Async', () => {
  test('async function() {}', () => {
    const result = babylon.parseExpression('async function() {}') as FunctionExpression
    expect(result.type).toEqual('FunctionExpression')
    expect(result.generator).toBeFalsy()
    expect(result.async).toBeTruthy()
    expect(result.params.length).toEqual(0)
    const { code } = generate(result)
    expect(code).toEqual('async function () {}')
  })

  test('async () => {}', () => {
    const result = babylon.parseExpression('async () => {}') as ArrowFunctionExpression
    expect(result.type).toEqual('ArrowFunctionExpression')
    expect(result.generator).toBeFalsy()
    expect(result.async).toBeTruthy()
    expect(result.params.length).toEqual(0)
    const { code } = generate(result)
    expect(code).toEqual('async () => {}')
  })
})

describe('Generator', () => {
  test('function * () {}', () => {
    const result = babylon.parseExpression('function * () {}') as FunctionExpression
    expect(result.type).toEqual('FunctionExpression')
    expect(result.generator).toBeTruthy()
    expect(result.async).toBeFalsy()
    expect(result.params.length).toEqual(0)
    const { code } = generate(result)
    expect(code).toEqual('function* () {}')
  })

  test('async function * () {}', () => {
    const result = babylon.parseExpression('async function * () {}', {
      plugins: [
        'asyncGenerators'
      ]
    }) as FunctionExpression
    expect(result.type).toEqual('FunctionExpression')
    expect(result.generator).toBeTruthy()
    expect(result.async).toBeTruthy()
    expect(result.params.length).toEqual(0)
    const { code } = generate(result)
    expect(code).toEqual('async function* () {}')
  })
})

describe('Object Method', () => {
  test('ObjectMethodName() {}', () => {
    const result = babylon.parseExpression('{' + 'ObjectMethodName() {}' + '}') as ObjectExpression
    expect(result.type).toEqual('ObjectExpression')
    expect(result.properties.length).toEqual(1)
    const method = result.properties[0] as ObjectMethod
    expect(method.type).toEqual('ObjectMethod')
    expect(method.generator).toBeFalsy()
    expect(method.async).toBeFalsy()
    expect(method.params.length).toEqual(0)
    const key = method.key as Identifier
    expect(key.type).toEqual('Identifier')
    expect(key.name).toEqual('ObjectMethodName')
    const { code } = generate(method)
    expect(code).toEqual('ObjectMethodName() {}')
  })

  test("['ObjectMethodName']() {}", () => {
    const result = babylon.parseExpression('{' + "['ObjectMethodName']() {}" + '}') as ObjectExpression
    expect(result.type).toEqual('ObjectExpression')
    expect(result.properties.length).toEqual(1)
    const method = result.properties[0] as ObjectMethod
    expect(method.type).toEqual('ObjectMethod')
    expect(method.generator).toBeFalsy()
    expect(method.async).toBeFalsy()
    expect(method.params.length).toEqual(0)
    const key = method.key as StringLiteral
    expect(key.type).toEqual('StringLiteral')
    expect(key.value).toEqual('ObjectMethodName')
    const { code } = generate(method)
    expect(code).toEqual("['ObjectMethodName']() {}")
  })

  test('[0]() {}', () => {
    const result = babylon.parseExpression('{' + '[0]() {}' + '}') as ObjectExpression
    expect(result.type).toEqual('ObjectExpression')
    expect(result.properties.length).toEqual(1)
    const method = result.properties[0] as ObjectMethod
    expect(method.type).toEqual('ObjectMethod')
    expect(method.generator).toBeFalsy()
    expect(method.async).toBeFalsy()
    expect(method.params.length).toEqual(0)
    const key = method.key as NumericLiteral
    expect(key.type).toEqual('NumericLiteral')
    expect(key.value).toEqual(0)
    const { code } = generate(method)
    expect(code).toEqual('[0]() {}')
  })

  test('async * ObjectMethodName() {}', () => {
    const result = babylon.parseExpression('{' + 'async * ObjectMethodName() {}' + '}', {
      plugins: [
        'asyncGenerators'
      ]
    }) as ObjectExpression
    expect(result.type).toEqual('ObjectExpression')
    expect(result.properties.length).toEqual(1)
    const method = result.properties[0] as ObjectMethod
    expect(method.type).toEqual('ObjectMethod')
    expect(method.generator).toBeTruthy()
    expect(method.async).toBeTruthy()
    expect(method.params.length).toEqual(0)
    const key = method.key as Identifier
    expect(key.type).toEqual('Identifier')
    expect(key.name).toEqual('ObjectMethodName')
    const { code } = generate(method)
    expect(code).toEqual('async *ObjectMethodName() {}')
  })

  test('async * [`ObjectMethodName`]() {}', () => {
    const result = babylon.parseExpression('{' + 'async * [`ObjectMethodName`]() {}' + '}', {
      plugins: [
        'asyncGenerators'
      ]
    }) as ObjectExpression
    expect(result.type).toEqual('ObjectExpression')
    expect(result.properties.length).toEqual(1)
    const method = result.properties[0] as ObjectMethod
    expect(method.type).toEqual('ObjectMethod')
    expect(method.generator).toBeTruthy()
    expect(method.async).toBeTruthy()
    expect(method.params.length).toEqual(0)
    const key = method.key as TemplateLiteral
    expect(key.type).toEqual('TemplateLiteral')
    expect(key.quasis[0].type).toEqual('TemplateElement')
    expect(key.quasis[0].value.cooked).toEqual('ObjectMethodName')
    const { code } = generate(method)
    expect(code).toEqual('async *[`ObjectMethodName`]() {}')
  })

  test('async * [0]() {}', () => {
    const result = babylon.parseExpression('{' + 'async * [0]() {}' + '}', {
      plugins: [
        'asyncGenerators'
      ]
    }) as ObjectExpression
    expect(result.type).toEqual('ObjectExpression')
    expect(result.properties.length).toEqual(1)
    const method = result.properties[0] as ObjectMethod
    expect(method.type).toEqual('ObjectMethod')
    expect(method.generator).toBeTruthy()
    expect(method.async).toBeTruthy()
    expect(method.params.length).toEqual(0)
    const key = method.key as NumericLiteral
    expect(key.type).toEqual('NumericLiteral')
    expect(key.value).toEqual(0)
    const { code } = generate(method)
    expect(code).toEqual('async *[0]() {}')
  })
})
